import uproot
import lhcb_rex.settings.globals as myGlobals
from rich.progress import Progress
import torch
from torch_geometric.data import Data
from torch_geometric.loader import DataLoader as geoDataLoader
from torch.utils.data import DataLoader
import pickle
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np
from rich import print
from torch_geometric.data import HeteroData

# import fast_vertex_quality_training.tools.globals as myGlobals
import lhcb_rex.data.loader as loader
import pandas as pd
import itertools


class StandardDataset(torch.utils.data.Dataset):
    """Dataset class for non-graph data."""

    def __init__(self, data, targets_graph, conditions_graph):
        self.graph_features_full = torch.tensor(
            data[targets_graph].values, dtype=torch.float
        )  # .to(myGlobals.device)
        self.graph_conditions_full = torch.tensor(
            data[conditions_graph].values, dtype=torch.float
        )  # .to(myGlobals.device)

        self.targets_graph = targets_graph
        self.conditions_graph = conditions_graph

    def __len__(self):
        return self.graph_features_full.shape[0]

    def __getitem__(self, idx):
        return {
            "targets": self.graph_features_full[idx],
            "conditions": self.graph_conditions_full[idx],
        }


class CustomData(Data):
    def __inc__(self, key, value, *args, **kwargs):
        if key == "hyperedge_index":
            # print(value)
            # quit()
            Nintermediates = int(torch.max(value[1]) + 1)
            # print(increment)
            # quit()
            return torch.tensor(
                [[self.num_nodes], [Nintermediates]]
            )  # Increment only the first row
        return super().__inc__(key, value, *args, **kwargs)


class CustomHeteroData(HeteroData):
    def __init__(self):
        super().__init__()

    def __inc__(self, key, value, store=None):
        """
        Customizes how to adjust the node indices for each batch.
        Returns the increment tensor to adjust node indices when batching.
        """

        # return super().__inc__(key, value, store)

        # if key == 'edge_index':
        #     return self['N_nodes']

        try:
            return super().__inc__(key, value, store)
        except Exception:
            # print(store.key)
            store = store["edge_index"]
            # print(self.num_nodes)
            # print(key, value, store, self)
            # quit()
            return super().__inc__(key, value, store)
            # print(store)
            # print(type(store))
            # print(store.size())
            # print(store['edge_index'].size())
            # quit()


class HeteroDataHandler:
    def __init__(
        self,
        particles_involved,
        intermediates,
        path,
        N=-1,
        entry_start=-1,
        mother_targets=[],
        intermediate_targets=[],
        track_targets=[],
        mother_conditions=[],
        intermediate_conditions=[],
        track_conditions=[],
        edge_conditions=[],
        graphify=False,
        inference=False,
        testing_frac=0.1,
        cut="",
        use_weights=True,
        splits=1,
        mother_N=3,
        intermediate_N=[],
        processID=-1,
        smearingnet=False,
        PIDnet=False,
        require_validation_variables=False,
    ):
        self.printed = False
        self.smearingnet = smearingnet
        self.PIDnet = PIDnet

        self.use_weights = use_weights

        self.first_row_printed = False

        self.particles_involved = particles_involved
        self.intermediates = intermediates
        self.N_daughters = len(self.particles_involved)

        self.mother_N = mother_N
        self.intermediate_N = intermediate_N

        self.mother_targets = mother_targets

        self.full_track_targets = []
        self.track_targets = track_targets
        for i in self.particles_involved:
            for ii in track_targets:
                self.full_track_targets.append(ii.replace("DAUGHTERN", f"DAUGHTER{i}"))

        self.full_track_conditions = []
        self.track_conditions = track_conditions
        for i in self.particles_involved:
            for ii in track_conditions:
                if "NMINUS1" in ii:
                    self.full_track_conditions.append(ii.replace("NMINUS1", f"{i - 1}"))
                elif "DAUGHTERN" in ii:
                    self.full_track_conditions.append(
                        ii.replace("DAUGHTERN", f"DAUGHTER{i}")
                    )
                else:
                    self.full_track_conditions.append(ii)

        self.edge_conditions = []
        self.len_edge_conditions = len(edge_conditions)
        if len(edge_conditions) > 0:
            for idx_i, i in enumerate(self.particles_involved):
                for idx_j, j in enumerate(self.particles_involved):
                    if idx_i != idx_j:
                        for ii in edge_conditions:
                            self.edge_conditions.append(
                                ii.replace(
                                    "DAUGHTERN_DAUGHTERN", f"DAUGHTER{i}_DAUGHTER{j}"
                                )
                            )

        self.mother_conditions = mother_conditions
        self.full_intermediate_targets = []
        self.full_intermediate_conditions = []
        self.intermediate_targets = intermediate_targets
        self.intermediate_conditions = intermediate_conditions
        self.intermediate_names = []
        for idx, i in enumerate(self.intermediates):
            for ii in intermediate_targets:
                if len(self.intermediates) == 1:
                    self.full_intermediate_targets.append(
                        ii.replace("MOTHER", "INTERMEDIATE")
                    )
                else:
                    self.full_intermediate_targets.append(
                        ii.replace("MOTHER", f"INTERMEDIATE{idx}")
                    )

            for ii in intermediate_conditions:
                if len(self.intermediates) == 1:
                    self.full_intermediate_conditions.append(
                        ii.replace("MOTHER", "INTERMEDIATE")
                    )
                else:
                    self.full_intermediate_conditions.append(
                        ii.replace("MOTHER", f"INTERMEDIATE{idx}")
                    )

            if len(self.intermediates) == 1:
                self.intermediate_names.append("INTERMEDIATE")
            else:
                self.intermediate_names.append(f"INTERMEDIATE{idx}")

        self.graphify = graphify

        print("\n track_targets:", self.full_track_targets)
        print("\n mother_targets:", self.mother_targets)
        print("\n intermediate_targets:", self.full_intermediate_targets)
        print("\n track_conditions:", self.full_track_conditions)
        print("\n mother_conditions:", self.mother_conditions)
        print("\n intermediate_conditions:", self.full_intermediate_conditions)
        print("\n edge_conditions:", self.edge_conditions)

        file = uproot.open(path)["DecayTree"]
        branches = (
            self.full_track_targets
            + self.mother_targets
            + self.full_intermediate_targets
            + self.full_track_conditions
            + self.mother_conditions
            + self.full_intermediate_conditions
            + self.edge_conditions
        )
        self.require_validation_variables = require_validation_variables
        if self.require_validation_variables:
            branches.extend(myGlobals.validation_variables)

        if self.use_weights:
            branches += ["weight"]

        if entry_start == -1:
            entry_start = None
        else:
            N += entry_start

        load_branches = branches
        print("loading events")
        print(f"file.num_entries: {file.num_entries}")

        if processID != -1:
            print(f"processID: {processID}")
            self.num_entries = file.num_entries
            print(f"num_entries, {self.num_entries}")
            self.num_entries_per_split = int(np.floor(self.num_entries / splits)) + 1
            start_entry = self.num_entries_per_split * processID
            end_entry = min(
                self.num_entries - 1, self.num_entries_per_split * (processID + 1)
            )
            if N != -1:
                end_entry = start_entry + N
            print(f"start_entry, {start_entry}")
            print(f"end_entry, {end_entry}")
        else:
            start_entry, end_entry = (
                entry_start,
                None if cut != "" or N <= 0 or N > file.num_entries else N,
            )

        if not self.smearingnet and not self.PIDnet:
            load_branches += ["fully_reco"]
        else:
            extra_PID_branches = [
                f"DAUGHTER{D + 1}_TRUEID" for D in range(self.N_daughters)
            ]
            load_branches += extra_PID_branches

        branches_to_array = [b for b in load_branches if b != "N_daughters"]
        branches_to_array = list(np.unique(branches_to_array))
        events = file.arrays(
            branches_to_array,
            library="pd",
            entry_start=start_entry,
            entry_stop=end_entry,
        )

        print("events loaded")

        print(events.shape, cut)
        events = events.loc[:, ~events.columns.str.contains("^Unnamed")]

        if cut != "" and processID == -1:
            events = events.query(cut)
            events = events.reset_index(drop=True)
            events = events.head(N)

        events["N_daughters"] = self.N_daughters
        branches += ["N_daughters"]

        self.data = loader.Dataset(particles_involved=particles_involved, filename=path)

        if inference:
            self.branches = branches
            self.data.fill(events, testing_frac=0.0)
            return
        self.data.fill(events, testing_frac=testing_frac)

        data_train = self.data.get_branches(branches, option="training")
        data_test = self.data.get_branches(branches, option="testing")

        # data_train = self.data.get_branches(branches, option="training", processed=False)
        # data_test = self.data.get_branches(branches, option="testing", processed=False)

        if self.smearingnet or self.PIDnet:
            self.extra_PID_branches = []
            for D in range(self.N_daughters):
                self.extra_PID_branches.append(f"DAUGHTER{D + 1}_TRUEID")
            data_train = data_train.drop(self.extra_PID_branches, axis=1)
            data_train_extra = self.data.get_branches(
                self.extra_PID_branches, option="training", processed=False
            )
            data_train = data_train.reset_index(drop=True)
            data_train_extra = data_train_extra.reset_index(drop=True)
            data_train = pd.concat((data_train, data_train_extra), axis=1)
            data_test = data_test.drop(self.extra_PID_branches, axis=1)
            data_test_extra = self.data.get_branches(
                self.extra_PID_branches, option="testing", processed=False
            )
            data_test = data_test.reset_index(drop=True)
            data_test_extra = data_test_extra.reset_index(drop=True)
            data_test = pd.concat((data_test, data_test_extra), axis=1)

        print(data_train)
        print(data_train.keys())
        # quit()

        if splits != 1 and processID == -1:
            print(
                f"Requested {splits} splits, exiting before making graphs, make graphs with make_split()"
            )
            self.data_train = data_train
            self.data_test = data_test

            self.data_train_N = self.data_train.shape[0]
            self.data_test_N = self.data_test.shape[0]
            self.N_train_per_split = int(np.floor(self.data_train_N / splits)) + 1
            self.N_test_per_split = int(np.floor(self.data_test_N / splits)) + 1

            return

        # data_train.to_pickle("./info_that_goes_into_graph.pkl")
        # data_train = self.data.get_branches(branches, option="training", processed=False)
        # data_train.to_pickle("./info_that_goes_into_graph_physical.pkl")
        # quit()
        # print(data_train["edge_angle_DAUGHTER1_DAUGHTER3"])
        # print(data_train["edge_angle_DAUGHTER2_DAUGHTER3"])
        # quit()

        # print(np.unique(data_train['DAUGHTER1_TRUEID']))
        # data_train_np = self.data.get_branches(branches, option="training", processed=False)
        # print(np.unique(data_train_np['DAUGHTER1_TRUEID']))

        # quit()
        if self.graphify:
            data_train_list = []
            data_train_list_weights = []
            data_train_list_full_reco = []
            with Progress() as progress:
                task = progress.add_task(
                    "Processing graphs (train)...", total=len(data_train)
                )

                for _, row in data_train.iterrows():
                    if self.smearingnet or self.PIDnet:
                        graph_i = self.create_data_from_row_smearingnet(row)
                    else:
                        # graph_i = self.create_data_from_row_uniquetracks(row)
                        graph_i = self.create_data_from_row(row)
                    if graph_i is not None:
                        data_train_list.append(graph_i)
                        if self.use_weights:
                            data_train_list_weights.append(row.weight)
                        if not self.smearingnet and not self.PIDnet:
                            data_train_list_full_reco.append(row.fully_reco)
                    progress.update(task, advance=1)
            # torch.save((data_train_list, data_train), "data.pkl")
            print(f"(data_train_list length) {len(data_train_list)}")
            data_test_list = []
            data_test_list_full_reco = []
            with Progress() as progress:
                task = progress.add_task(
                    "Processing graphs (test)...", total=len(data_test)
                )

                for _, row in data_test.iterrows():
                    if self.smearingnet or self.PIDnet:
                        graph_i = self.create_data_from_row_smearingnet(row)
                    else:
                        # graph_i = self.create_data_from_row_uniquetracks(row)
                        graph_i = self.create_data_from_row(row)
                    if graph_i is not None:
                        data_test_list.append(graph_i)
                        if not self.smearingnet and not self.PIDnet:
                            data_test_list_full_reco.append(row.fully_reco)
                    progress.update(task, advance=1)
            # torch.save((data_test_list, data_test), "data_test.pkl")

            self.data_train = data_train_list
            self.data_test = data_test_list
            self.data_train_list_full_reco = data_train_list_full_reco
            self.data_test_list_full_reco = data_test_list_full_reco
            if self.use_weights:
                self.data_train_list_weights = data_train_list_weights

        else:
            self.data_train = data_train
            self.data_test = data_test
            self.data_test_raw = data_test

    def make_split(self, split):
        start = self.N_train_per_split * split
        end = self.N_train_per_split * (split + 1)
        if end >= self.data_train_N:
            end = self.data_train_N - 1
        print(split, start, end, f"N: {end - start}")
        data_train_i = self.data_train.iloc[int(start) : int(end)]

        start = self.N_test_per_split * split
        end = self.N_test_per_split * (split + 1)
        if end >= self.data_test_N:
            end = self.data_test_N - 1
        print(split, start, end, f"N: {end - start}")
        data_test_i = self.data_test.iloc[int(start) : int(end)]

        data_train_list = []
        data_train_list_weights = []
        with Progress() as progress:
            task = progress.add_task(
                "Processing graphs (train)...", total=len(data_train_i)
            )

            for _, row in data_train_i.iterrows():
                if self.smearingnet or self.PIDnet:
                    graph_i = self.create_data_from_row_smearingnet(row)
                else:
                    # graph_i = self.create_data_from_row_uniquetracks(row)
                    graph_i = self.create_data_from_row(row)
                if graph_i is not None:
                    data_train_list.append(graph_i)
                    if self.use_weights:
                        data_train_list_weights.append(row.weight)
                progress.update(task, advance=1)

        data_test_list = []
        with Progress() as progress:
            task = progress.add_task(
                "Processing graphs (test)...", total=len(data_test_i)
            )

            for _, row in data_test_i.iterrows():
                if self.smearingnet or self.PIDnet:
                    graph_i = self.create_data_from_row_smearingnet(row)
                else:
                    # graph_i = self.create_data_from_row_uniquetracks(row)
                    graph_i = self.create_data_from_row(row)
                if graph_i is not None:
                    data_test_list.append(graph_i)
                progress.update(task, advance=1)

        self.data_train_i = data_train_list
        self.data_test_i = data_test_list
        if self.use_weights:
            self.data_train_list_weights_i = data_train_list_weights

    def plot(
        self,
        filename,
        variables=None,
        overlay=None,
        bins=50,
        label=None,
    ):
        if overlay:
            if not isinstance(overlay, list):
                overlay = [overlay]

        if variables is None:
            variables = list(self.data.all_data["physical"].keys())

        with PdfPages(filename) as pdf:
            for variable in variables:
                try:
                    plt.figure(figsize=(10, 8))

                    plt.subplot(2, 2, 1)
                    plt.title(variable)
                    if overlay:
                        info = [self.data.all_data["physical"][variable]]
                        for overlay_i in overlay:
                            info.append(overlay_i.data.all_data["physical"][variable])
                        plt.hist(
                            info, bins=bins, density=True, histtype="step", label=label
                        )
                    else:
                        plt.hist(
                            self.data.all_data["physical"][variable],
                            bins=bins,
                            density=True,
                            histtype="step",
                            label=label,
                        )

                    if label:
                        plt.legend()

                    plt.subplot(2, 2, 2)
                    plt.title(f"{variable} processed")
                    if overlay:
                        info = [self.data.all_data["processed"][variable]]
                        for overlay_i in overlay:
                            info.append(overlay_i.data.all_data["processed"][variable])
                        plt.hist(info, bins=bins, density=True, histtype="step")
                    else:
                        plt.hist(
                            self.data.all_data["processed"][variable],
                            bins=bins,
                            density=True,
                            histtype="step",
                            range=[-1, 1],
                        )

                    plt.subplot(2, 2, 3)
                    if overlay:
                        info = [self.data.all_data["physical"][variable]]
                        for overlay_i in overlay:
                            info.append(overlay_i.all_data["physical"][variable])
                        plt.hist(info, bins=bins, density=True, histtype="step")
                    else:
                        plt.hist(
                            self.data.all_data["physical"][variable],
                            bins=bins,
                            density=True,
                            histtype="step",
                        )
                    plt.yscale("log")

                    plt.subplot(2, 2, 4)
                    if overlay:
                        info = [self.data.all_data["processed"][variable]]
                        for overlay_i in overlay:
                            info.append(overlay_i.data.all_data["processed"][variable])
                        plt.hist(info, bins=bins, density=True, histtype="step")
                    else:
                        plt.hist(
                            self.data.all_data["processed"][variable],
                            bins=bins,
                            density=True,
                            histtype="step",
                            range=[-1, 1],
                        )
                    plt.yscale("log")

                    pdf.savefig(bbox_inches="tight")
                    plt.close()

                except Exception:
                    pdf.savefig(bbox_inches="tight")
                    plt.close()
                    pass

    def save_graphs(self, save_loc):
        data = {
            "data_train": self.data_train,
            "data_test": self.data_test,
        }

        # Save the dictionary to a .pkl file
        with open(save_loc, "wb") as f:
            pickle.dump(data, f)

    def load_graphs(self, save_loc):
        # Load the data from the specified .pkl file
        with open(save_loc, "rb") as f:
            data = pickle.load(f)

        # Assign loaded data to the corresponding attributes
        self.data_train = data.get("data_train")
        self.data_test = data.get("data_test")

    # def create_data_from_row_uniquetracks(self, row):
    #     nan_cols = row.index[row.isna()].tolist()  # Get column names with NaNs
    #     if nan_cols:  # If there are any NaNs
    #         print(f"Row has NaNs in columns: {nan_cols}")
    #         return None

    #     data = CustomHeteroData()  # allows for edge_index to be list

    #     particle_types = ["11", "13", "211", "321"]

    #     track_targets = torch.tensor(
    #         row[self.full_track_targets].values, dtype=torch.float
    #     )
    #     track_targets = track_targets.view(-1, len(self.track_targets))
    #     track_conditions = torch.tensor(
    #         row[self.full_track_conditions].values, dtype=torch.float
    #     )
    #     track_conditions = track_conditions.view(-1, len(self.track_conditions))

    #     for particle_type in particle_types:
    #         data[particle_type].x = torch.empty(
    #             (0, track_targets.shape[-1]), dtype=torch.long
    #         )
    #         data[f"{particle_type}_conditions"].x = torch.empty(
    #             (0, track_conditions.shape[-1]), dtype=torch.long
    #         )
    #         data[f"{particle_type}_trackidx"].x = torch.empty((0), dtype=torch.long)

    #     track_PIDs = np.abs(np.asarray(row[self.extra_PID_branches].values)).astype(int)

    #     for track_global_index, track_PID in enumerate(track_PIDs):
    #         data[str(track_PID)].x = torch.cat(
    #             (
    #                 data[str(track_PID)].x,
    #                 torch.unsqueeze(track_targets[track_global_index], dim=0),
    #             ),
    #             dim=0,
    #         )
    #         data[f"{str(track_PID)}_conditions"].x = torch.cat(
    #             (
    #                 data[f"{str(track_PID)}_conditions"].x,
    #                 torch.unsqueeze(track_conditions[track_global_index], dim=0),
    #             ),
    #             dim=0,
    #         )
    #         track_global_index = torch.tensor([track_global_index], dtype=torch.long)
    #         data[f"{str(track_PID)}_trackidx"].x = torch.cat(
    #             (data[f"{str(track_PID)}_trackidx"].x, track_global_index), dim=0
    #         )

    #     print(track_PIDs)
    #     print(self.extra_PID_branches)

    #     if len(self.full_intermediate_targets) > 0:
    #         intermediate_targets = torch.tensor(
    #             row[self.full_intermediate_targets].values, dtype=torch.float
    #         )
    #         intermediate_targets = intermediate_targets.view(
    #             -1, len(self.intermediate_targets)
    #         )
    #         data[
    #             "intermediate"
    #         ].x = intermediate_targets  # [num_intermediates, intermediate_targets]
    #         intermediate_conditions = torch.tensor(
    #             row[self.full_intermediate_conditions].values, dtype=torch.float
    #         )
    #         intermediate_conditions = intermediate_conditions.view(
    #             -1, len(self.intermediate_conditions)
    #         )
    #         if "N_daughters" in self.intermediate_conditions:
    #             idx = self.intermediate_conditions.index("N_daughters")
    #             for int_idx, N in enumerate(self.intermediate_N):
    #                 intermediate_conditions[int_idx][idx] = N
    #         data[
    #             "intermediate_conditions"
    #         ].x = (
    #             intermediate_conditions  # [num_intermediates, intermediate_conditions]
    #         )

    #     if len(self.mother_targets) > 0:
    #         mother_targets = torch.tensor(
    #             row[self.mother_targets].values, dtype=torch.float
    #         )
    #         mother_targets = mother_targets.view(-1, len(self.mother_targets))
    #         data["mother"].x = mother_targets  # [num_mothers, mother_targets]
    #         mother_conditions = torch.tensor(
    #             row[self.mother_conditions].values, dtype=torch.float
    #         )
    #         mother_conditions = mother_conditions.view(-1, len(self.mother_conditions))
    #         if "N_daughters" in self.mother_conditions:
    #             idx = self.mother_conditions.index("N_daughters")
    #             mother_conditions[0][idx] = self.mother_N
    #         data[
    #             "mother_conditions"
    #         ].x = mother_conditions  # [num_mothers, mother_conditions]

    #     if self.require_validation_variables:
    #         validation_variables = torch.tensor(
    #                 row[myGlobals.validation_variables].values, dtype=torch.float
    #             )
    #         data["validation_variables"].x = validation_variables

    #     # edges_by_index = list(itertools.product(range(len(track_PIDs)), repeat=2))

    #     index_by_appearence = []
    #     index_counters = {11: 0, 13: 0, 211: 0, 321: 0}
    #     for item in track_PIDs:
    #         index_by_appearence.append(index_counters[item])
    #         index_counters[item] += 1

    #     print(index_counters)

    #     data["intermediate", "to", "intermediate"].edge_index = (
    #         torch.tensor([[0, 0]], dtype=torch.long).t().contiguous()
    #     )
    #     data["mother", "to", "mother"].edge_index = (
    #         torch.tensor([[0, 0]], dtype=torch.long).t().contiguous()
    #     )
    #     data["intermediate", "to", "mother"].edge_index = [
    #         torch.tensor([[0, 0]], dtype=torch.long).t().contiguous()
    #     ]
    #     data["mother", "to", "intermediate"].edge_index = [
    #             torch.tensor([[0, 0]], dtype=torch.long).t().contiguous()
    #         ]

    #     for node_type in particle_types:
    #         data[node_type, "to", "intermediate"].edge_index = torch.empty(
    #             (2, 0), dtype=torch.long
    #         )
    #         data[node_type, "to", "mother"].edge_index = torch.empty(
    #             (2, 0), dtype=torch.long
    #         )

    #         # data[pair[0], "to", pair[1]].edge_attr = torch.empty(
    #         #     (0, self.len_edge_conditions), dtype=torch.long
    #         # )

    #     print('here')
    #     quit()

    #     track_targets = torch.tensor(
    #         row[self.full_track_targets].values, dtype=torch.float
    #     )
    #     track_targets = track_targets.view(-1, len(self.track_targets))
    #     data["track"].x = track_targets  # [num_tracks, track_targets]
    #     track_conditions = torch.tensor(
    #         row[self.full_track_conditions].values, dtype=torch.float
    #     )
    #     track_conditions = track_conditions.view(-1, len(self.track_conditions))
    #     data["track_conditions"].x = track_conditions  # [num_tracks, track_conditions]

    #     if len(self.full_intermediate_targets) > 0:
    #         intermediate_targets = torch.tensor(
    #             row[self.full_intermediate_targets].values, dtype=torch.float
    #         )
    #         intermediate_targets = intermediate_targets.view(
    #             -1, len(self.intermediate_targets)
    #         )
    #         data[
    #             "intermediate"
    #         ].x = intermediate_targets  # [num_intermediates, intermediate_targets]
    #         intermediate_conditions = torch.tensor(
    #             row[self.full_intermediate_conditions].values, dtype=torch.float
    #         )
    #         intermediate_conditions = intermediate_conditions.view(
    #             -1, len(self.intermediate_conditions)
    #         )
    #         if "N_daughters" in self.intermediate_conditions:
    #             idx = self.intermediate_conditions.index("N_daughters")
    #             for int_idx, N in enumerate(self.intermediate_N):
    #                 intermediate_conditions[int_idx][idx] = N
    #         data[
    #             "intermediate_conditions"
    #         ].x = (
    #             intermediate_conditions  # [num_intermediates, intermediate_conditions]
    #         )

    #     if len(self.mother_targets) > 0:
    #         mother_targets = torch.tensor(
    #             row[self.mother_targets].values, dtype=torch.float
    #         )
    #         mother_targets = mother_targets.view(-1, len(self.mother_targets))
    #         data["mother"].x = mother_targets  # [num_mothers, mother_targets]
    #         mother_conditions = torch.tensor(
    #             row[self.mother_conditions].values, dtype=torch.float
    #         )
    #         mother_conditions = mother_conditions.view(-1, len(self.mother_conditions))
    #         if "N_daughters" in self.mother_conditions:
    #             idx = self.mother_conditions.index("N_daughters")
    #             mother_conditions[0][idx] = self.mother_N
    #         data[
    #             "mother_conditions"
    #         ].x = mother_conditions  # [num_mothers, mother_conditions]

    #     if self.require_validation_variables:
    #         validation_variables = torch.tensor(
    #                 row[myGlobals.validation_variables].values, dtype=torch.float
    #             )
    #         data["validation_variables"].x = validation_variables

    #     node_indexes = {}
    #     node_indexes["tracks"] = []
    #     node_indexes["intermediates"] = []
    #     node_indexes["mother"] = []
    #     node_index = 0
    #     for track in range(track_targets.shape[0]):
    #         node_indexes["tracks"].append(node_index)
    #         node_index += 1
    #     if len(self.full_intermediate_targets) > 0:
    #         for intermediate in range(intermediate_targets.shape[0]):
    #             node_indexes["intermediates"].append(node_index)
    #             node_index += 1
    #     if len(self.mother_targets) > 0:
    #         node_indexes["mother"].append(node_index)

    #     empty = torch.tensor([[], []], dtype=torch.long)

    #     idx_straight_to_mother = 0
    #     idx_int_track_i = 1
    #     idx_int_track_j = 2

    #     # this is shite, need to generalise later... this is fine for initial testing
    #     if len(node_indexes["intermediates"]) > 0:
    #         # self edges
    #         data["track", "to", "track"].edge_index = (
    #             torch.tensor([[0, 0], [1, 1], [2, 2]], dtype=torch.long)
    #             .t()
    #             .contiguous()
    #         )
    #         data["intermediate", "to", "intermediate"].edge_index = (
    #             torch.tensor([[0, 0]], dtype=torch.long).t().contiguous()
    #         )
    #         data["mother", "to", "mother"].edge_index = (
    #             torch.tensor([[0, 0]], dtype=torch.long).t().contiguous()
    #         )

    #         # forward edges
    #         data["track", "to", "intermediate"].edge_index = [
    #             torch.tensor([[idx_int_track_i, 0], [idx_int_track_j, 0]], dtype=torch.long).t().contiguous()
    #         ]
    #         data["track", "skip", "intermediate"].edge_index = [
    #             torch.tensor([[idx_straight_to_mother, 0]], dtype=torch.long).t().contiguous()
    #         ]

    #         data["intermediate", "to", "mother"].edge_index = [
    #             torch.tensor([[0, 0]], dtype=torch.long).t().contiguous()
    #         ]

    #         data["track", "to", "mother"].edge_index = [
    #             torch.tensor([[idx_straight_to_mother, 0]], dtype=torch.long).t().contiguous()
    #         ]
    #         data["track", "skip", "mother"].edge_index = [
    #             torch.tensor([[idx_int_track_i, 0], [idx_int_track_j, 0]], dtype=torch.long).t().contiguous()
    #         ]

    #         # backward edges
    #         data["intermediate", "to", "track"].edge_index = [
    #             torch.tensor([[0, idx_int_track_i], [0, idx_int_track_j]], dtype=torch.long).t().contiguous()
    #         ]
    #         data["intermediate", "skip", "track"].edge_index = [
    #             torch.tensor([[0, idx_straight_to_mother]], dtype=torch.long).t().contiguous()
    #         ]

    #         data["mother", "to", "intermediate"].edge_index = [
    #             torch.tensor([[0, 0]], dtype=torch.long).t().contiguous()
    #         ]

    #         data["mother", "to", "track"].edge_index = [
    #             torch.tensor([[0, idx_straight_to_mother]], dtype=torch.long).t().contiguous()
    #         ]
    #         data["mother", "skip", "track"].edge_index = [
    #             torch.tensor([[0, idx_int_track_i], [0, idx_int_track_j]], dtype=torch.long).t().contiguous()
    #         ]

    #     else:
    #         data["track", "to", "track"].edge_index = (
    #             torch.tensor([[0, 0], [1, 1], [2, 2]], dtype=torch.long)
    #             .t()
    #             .contiguous()
    #         )
    #         data["intermediate", "to", "intermediate"].edge_index = empty
    #         data["mother", "to", "mother"].edge_index = (
    #             torch.tensor([[0, 0]], dtype=torch.long).t().contiguous()
    #         )

    #         data["track", "to", "intermediate"].edge_index = [empty]
    #         data["intermediate", "to", "mother"].edge_index = [empty]
    #         data["track", "to", "mother"].edge_index = [
    #             torch.tensor([[0, 0], [1, 0], [2, 0]], dtype=torch.long)
    #             .t()
    #             .contiguous()
    #         ]

    #         data["intermediate", "to", "track"].edge_index = [empty]
    #         data["mother", "to", "intermediate"].edge_index = [empty]
    #         data["mother", "to", "track"].edge_index = [
    #             torch.tensor([[0, 0], [0, 1], [0, 2]], dtype=torch.long)
    #             .t()
    #             .contiguous()
    #         ]

    #     return data

    def create_data_from_row(self, row):
        nan_cols = row.index[row.isna()].tolist()  # Get column names with NaNs
        if nan_cols:  # If there are any NaNs
            print(f"Row has NaNs in columns: {nan_cols}")
            return None

        data = CustomHeteroData()  # allows for edge_index to be list

        track_targets = torch.tensor(
            row[self.full_track_targets].values, dtype=torch.float
        )
        track_targets = track_targets.view(-1, len(self.track_targets))
        data["track"].x = track_targets  # [num_tracks, track_targets]
        track_conditions = torch.tensor(
            row[self.full_track_conditions].values, dtype=torch.float
        )
        track_conditions = track_conditions.view(-1, len(self.track_conditions))
        data["track_conditions"].x = track_conditions  # [num_tracks, track_conditions]

        if len(self.full_intermediate_targets) > 0:
            intermediate_targets = torch.tensor(
                row[self.full_intermediate_targets].values, dtype=torch.float
            )
            intermediate_targets = intermediate_targets.view(
                -1, len(self.intermediate_targets)
            )
            data[
                "intermediate"
            ].x = intermediate_targets  # [num_intermediates, intermediate_targets]
            intermediate_conditions = torch.tensor(
                row[self.full_intermediate_conditions].values, dtype=torch.float
            )
            intermediate_conditions = intermediate_conditions.view(
                -1, len(self.intermediate_conditions)
            )
            if "N_daughters" in self.intermediate_conditions:
                idx = self.intermediate_conditions.index("N_daughters")
                for int_idx, N in enumerate(self.intermediate_N):
                    intermediate_conditions[int_idx][idx] = N
            data[
                "intermediate_conditions"
            ].x = (
                intermediate_conditions  # [num_intermediates, intermediate_conditions]
            )

        if len(self.mother_targets) > 0:
            mother_targets = torch.tensor(
                row[self.mother_targets].values, dtype=torch.float
            )
            mother_targets = mother_targets.view(-1, len(self.mother_targets))
            data["mother"].x = mother_targets  # [num_mothers, mother_targets]
            mother_conditions = torch.tensor(
                row[self.mother_conditions].values, dtype=torch.float
            )
            mother_conditions = mother_conditions.view(-1, len(self.mother_conditions))
            if "N_daughters" in self.mother_conditions:
                idx = self.mother_conditions.index("N_daughters")
                mother_conditions[0][idx] = self.mother_N
            data[
                "mother_conditions"
            ].x = mother_conditions  # [num_mothers, mother_conditions]

        if self.require_validation_variables:
            validation_variables = torch.tensor(
                row[myGlobals.validation_variables].values, dtype=torch.float
            )
            data["validation_variables"].x = validation_variables

        node_indexes = {}
        node_indexes["tracks"] = []
        node_indexes["intermediates"] = []
        node_indexes["mother"] = []
        node_index = 0
        for track in range(track_targets.shape[0]):
            node_indexes["tracks"].append(node_index)
            node_index += 1
        if len(self.full_intermediate_targets) > 0:
            for intermediate in range(intermediate_targets.shape[0]):
                node_indexes["intermediates"].append(node_index)
                node_index += 1
        if len(self.mother_targets) > 0:
            node_indexes["mother"].append(node_index)

        empty = torch.tensor([[], []], dtype=torch.long)

        idx_straight_to_mother = 0
        idx_int_track_i = 1
        idx_int_track_j = 2

        # this is shite, need to generalise later... this is fine for initial testing
        if len(node_indexes["intermediates"]) > 0:
            # self edges
            data["track", "to", "track"].edge_index = (
                torch.tensor([[0, 0], [1, 1], [2, 2]], dtype=torch.long)
                .t()
                .contiguous()
            )
            data["intermediate", "to", "intermediate"].edge_index = (
                torch.tensor([[0, 0]], dtype=torch.long).t().contiguous()
            )
            data["mother", "to", "mother"].edge_index = (
                torch.tensor([[0, 0]], dtype=torch.long).t().contiguous()
            )

            # forward edges
            data["track", "to", "intermediate"].edge_index = [
                torch.tensor(
                    [[idx_int_track_i, 0], [idx_int_track_j, 0]], dtype=torch.long
                )
                .t()
                .contiguous(),
                empty,
            ]
            # data["track", "skip", "intermediate"].edge_index = [
            #     torch.tensor([[idx_straight_to_mother, 0]], dtype=torch.long).t().contiguous(), empty
            # ]

            data["intermediate", "to", "mother"].edge_index = [
                empty,
                torch.tensor([[0, 0]], dtype=torch.long).t().contiguous(),
            ]

            data["track", "to", "mother"].edge_index = [
                empty,
                torch.tensor([[idx_straight_to_mother, 0]], dtype=torch.long)
                .t()
                .contiguous(),
            ]
            # data["track", "skip", "mother"].edge_index = [
            #     empty, torch.tensor([[idx_int_track_i, 0], [idx_int_track_j, 0]], dtype=torch.long).t().contiguous()
            # ]

            # backward edges
            data["intermediate", "to", "track"].edge_index = [
                empty,
                torch.tensor(
                    [[0, idx_int_track_i], [0, idx_int_track_j]], dtype=torch.long
                )
                .t()
                .contiguous(),
            ]
            # data["intermediate", "skip", "track"].edge_index = [
            #     empty, torch.tensor([[0, idx_straight_to_mother]], dtype=torch.long).t().contiguous()
            # ]

            data["mother", "to", "intermediate"].edge_index = [
                torch.tensor([[0, 0]], dtype=torch.long).t().contiguous(),
                empty,
            ]

            data["mother", "to", "track"].edge_index = [
                torch.tensor([[0, idx_straight_to_mother]], dtype=torch.long)
                .t()
                .contiguous(),
                empty,
            ]
            # data["mother", "skip", "track"].edge_index = [
            #     empty, torch.tensor([[0, idx_int_track_i], [0, idx_int_track_j]], dtype=torch.long).t().contiguous()
            # ]

        else:
            data["track", "to", "track"].edge_index = (
                torch.tensor([[0, 0], [1, 1], [2, 2]], dtype=torch.long)
                .t()
                .contiguous()
            )
            data["intermediate", "to", "intermediate"].edge_index = empty
            data["mother", "to", "mother"].edge_index = (
                torch.tensor([[0, 0]], dtype=torch.long).t().contiguous()
            )

            data["track", "to", "intermediate"].edge_index = [empty]
            data["intermediate", "to", "mother"].edge_index = [empty]
            data["track", "to", "mother"].edge_index = [
                torch.tensor([[0, 0], [1, 0], [2, 0]], dtype=torch.long)
                .t()
                .contiguous()
            ]

            data["intermediate", "to", "track"].edge_index = [empty]
            data["mother", "to", "intermediate"].edge_index = [empty]
            data["mother", "to", "track"].edge_index = [
                torch.tensor([[0, 0], [0, 1], [0, 2]], dtype=torch.long)
                .t()
                .contiguous()
            ]

        return data

    def create_data_from_row_smearingnet(self, row):
        # print('\n\n\n')
        # print(row)
        # print(self.edge_conditions)
        # quit()

        particle_types = ["11", "13", "211", "321"]

        nan_cols = row.index[row.isna()].tolist()  # Get column names with NaNs
        if nan_cols:  # If there are any NaNs
            print(f"Row has NaNs in columns: {nan_cols}")
            return None

        data = CustomHeteroData()  # allows for edge_index to be list

        track_targets = torch.tensor(
            row[self.full_track_targets].values, dtype=torch.float
        )
        track_targets = track_targets.view(-1, len(self.track_targets))
        track_conditions = torch.tensor(
            row[self.full_track_conditions].values, dtype=torch.float
        )
        track_conditions = track_conditions.view(-1, len(self.track_conditions))

        for particle_type in particle_types:
            data[particle_type].x = torch.empty(
                (0, track_targets.shape[-1]), dtype=torch.long
            )
            data[f"{particle_type}_conditions"].x = torch.empty(
                (0, track_conditions.shape[-1]), dtype=torch.long
            )
            data[f"{particle_type}_trackidx"].x = torch.empty((0), dtype=torch.long)

        track_PIDs = np.abs(np.asarray(row[self.extra_PID_branches].values)).astype(int)

        for track_global_index, track_PID in enumerate(track_PIDs):
            data[str(track_PID)].x = torch.cat(
                (
                    data[str(track_PID)].x,
                    torch.unsqueeze(track_targets[track_global_index], dim=0),
                ),
                dim=0,
            )
            data[f"{str(track_PID)}_conditions"].x = torch.cat(
                (
                    data[f"{str(track_PID)}_conditions"].x,
                    torch.unsqueeze(track_conditions[track_global_index], dim=0),
                ),
                dim=0,
            )
            track_global_index = torch.tensor([track_global_index], dtype=torch.long)
            data[f"{str(track_PID)}_trackidx"].x = torch.cat(
                (data[f"{str(track_PID)}_trackidx"].x, track_global_index), dim=0
            )

        ordered_pairs = list(itertools.product(particle_types, repeat=2))

        for pair in ordered_pairs:
            data[pair[0], "to", pair[1]].edge_index = torch.empty(
                (2, 0), dtype=torch.long
            )
            data[pair[0], "to", pair[1]].edge_attr = torch.empty(
                (0, self.len_edge_conditions), dtype=torch.long
            )

        edges_by_index = list(itertools.product(range(len(track_PIDs)), repeat=2))

        index_by_appearence = []
        index_counters = {11: 0, 13: 0, 211: 0, 321: 0}
        for item in track_PIDs:
            index_by_appearence.append(index_counters[item])
            index_counters[item] += 1

        for edge_by_index in edges_by_index:
            origin = str(track_PIDs[edge_by_index[0]])
            destination = str(track_PIDs[edge_by_index[1]])
            origin_idx = index_by_appearence[edge_by_index[0]]
            destination_idx = index_by_appearence[edge_by_index[1]]

            if origin != destination or origin_idx != destination_idx:
                edge_i = torch.tensor(
                    [[origin_idx, destination_idx]], dtype=torch.long
                ).t()
                data[origin, "to", destination].edge_index = torch.cat(
                    (data[origin, "to", destination].edge_index, edge_i), dim=1
                )

                edge_attr_i = torch.tensor(
                    [
                        row[
                            f"edge_angle_DAUGHTER{edge_by_index[0] + 1}_DAUGHTER{edge_by_index[1] + 1}_TRUE"
                        ]
                    ],
                    dtype=torch.float,
                ).unsqueeze(0)
                data[origin, "to", destination].edge_attr = torch.cat(
                    (data[origin, "to", destination].edge_attr, edge_attr_i), dim=0
                )

        return data

    def create_graphs(
        self,
        targets_graph,
        targets_node,
        conditions_graph,
        conditions_node,
        N_daughters_override=None,
        fully_reco_override=None,
    ):  # used by inference.py
        self.targets_graph = targets_graph
        self.targets_node = targets_node
        self.conditions_graph = conditions_graph
        self.conditions_node = conditions_node

        branches = (
            self.targets_graph
            + self.targets_node
            + self.conditions_graph
            + self.conditions_node
        )

        data_test = self.data.get_branches(branches)

        if N_daughters_override is not None:
            data_test["N_daughters"] = N_daughters_override
        if fully_reco_override is not None:
            data_test["fully_reco"] = fully_reco_override

        data_test_list = []
        with Progress() as progress:
            task = progress.add_task(
                "Processing graphs (test)...", total=len(data_test)
            )

            for _, row in data_test.iterrows():
                if self.smearingnet:
                    data_test_list.append(self.create_data_from_row_smearingnet(row))
                else:
                    # data_test_list.append(self.create_data_from_row_uniquetracks(row))
                    data_test_list.append(self.create_data_from_row(row))
                progress.update(task, advance=1)

        self.data_test = data_test_list
        self.data_test_raw = data_test

        return len(data_test_list)

    def get_loaders(
        self, batch_size, only_test=False, graphify=False
    ):  # used by inference.py
        if graphify:
            self.graphify = graphify

        if self.graphify:
            if not only_test:
                loader_train = geoDataLoader(
                    self.data_train, batch_size=batch_size, shuffle=True
                )
            test_batch_size = len(self.data_test)
            loader_test = geoDataLoader(
                self.data_test, batch_size=test_batch_size, shuffle=False
            )
        else:
            if not only_test:
                train_dataset = StandardDataset(
                    self.data_train, self.targets_graph, self.conditions_graph
                )
                loader_train = DataLoader(
                    train_dataset, batch_size=batch_size, shuffle=True
                )

            test_dataset = StandardDataset(
                self.data_test, self.targets_graph, self.conditions_graph
            )
            test_batch_size = len(test_dataset)
            loader_test = DataLoader(
                test_dataset, batch_size=test_batch_size, shuffle=False
            )
        if only_test:
            return loader_test
        else:
            return loader_train, loader_test, self.data_test_raw, test_batch_size


class DataHandler:
    def __init__(
        self,
        particles_involved,
        path,
        N=-1,
        entry_start=-1,
        targets_graph=[],
        targets_node=[],
        conditions_graph=[],
        conditions_node=[],
        conditions_edge=[],
        graphify=False,
        inference=False,
        testing_frac=0.1,
        cut="",
        use_weights=True,
    ):
        self.use_weights = use_weights

        self.first_row_printed = False

        self.particles_involved = particles_involved
        self.N_daughters = len(self.particles_involved)
        self.targets_graph = targets_graph

        self.targets_node = []
        for i in self.particles_involved:
            for ii in targets_node:
                self.targets_node.append(ii.replace("DAUGHTERN", f"DAUGHTER{i}"))

        self.conditions_node = []
        for i in self.particles_involved:
            for ii in conditions_node:
                if "NMINUS1" in ii:
                    self.conditions_node.append(ii.replace("NMINUS1", f"{i - 1}"))
                elif "DAUGHTERN" in ii:
                    self.conditions_node.append(ii.replace("DAUGHTERN", f"DAUGHTER{i}"))
                else:
                    self.conditions_node.append(ii)

        self.conditions_edge = []
        for idx_i, i in enumerate(self.particles_involved):
            for idx_j, j in enumerate(self.particles_involved):
                if idx_i != idx_j:
                    for ii in conditions_edge:
                        self.conditions_edge.append(
                            ii.replace(
                                "DAUGHTERN_DAUGHTERN", f"DAUGHTER{i}_DAUGHTER{j}"
                            )
                        )

        self.conditions_graph = conditions_graph
        self.graphify = graphify

        print("\n targets_graph:", self.targets_graph)
        print("\n targets_node:", self.targets_node)
        print("\n conditions_graph:", self.conditions_graph)
        print("\n conditions_node:", self.conditions_node)
        print("\n conditions_edge:", self.conditions_edge)

        file = uproot.open(path)["DecayTree"]
        branches = (
            self.targets_graph
            + self.targets_node
            + self.conditions_graph
            + self.conditions_node
            + self.conditions_edge
        )
        if self.use_weights:
            branches += ["weight"]

        if entry_start == -1:
            entry_start = None
        else:
            N += entry_start

        load_branches = [branch for branch in file.keys() if "_COV_" not in branch]
        print("loading events")
        if cut != "" or N <= 0 or N > file.num_entries:
            events = file.arrays(load_branches, library="pd", entry_start=entry_start)
        else:
            events = file.arrays(
                load_branches, library="pd", entry_start=entry_start, entry_stop=N
            )
        print("events loaded")

        events = events.loc[:, ~events.columns.str.contains("^Unnamed")]

        if cut != "":
            events = events.query(cut)
            events = events.reset_index(drop=True)
            events = events.head(N)

        events["N_daughters"] = self.N_daughters

        self.data = loader.Dataset(particles_involved=particles_involved, filename=path)

        if inference:
            self.data.fill(events, testing_frac=0.0)
            return
        self.data.fill(events, testing_frac=testing_frac)

        data_train = self.data.get_branches(branches, option="training")
        data_test = self.data.get_branches(branches, option="testing")

        print(data_train)
        print(data_train.keys())
        # data_train.to_pickle("./info_that_goes_into_graph.pkl")
        # data_train = self.data.get_branches(branches, option="training", processed=False)
        # data_train.to_pickle("./info_that_goes_into_graph_physical.pkl")
        # quit()
        # print(data_train["edge_angle_DAUGHTER1_DAUGHTER3"])
        # print(data_train["edge_angle_DAUGHTER2_DAUGHTER3"])
        # quit()

        # print(np.unique(data_train['DAUGHTER1_TRUEID']))
        # data_train_np = self.data.get_branches(branches, option="training", processed=False)
        # print(np.unique(data_train_np['DAUGHTER1_TRUEID']))

        # quit()
        if self.graphify:
            data_train_list = []
            data_train_list_weights = []
            with Progress() as progress:
                task = progress.add_task(
                    "Processing graphs (train)...", total=len(data_train)
                )

                for _, row in data_train.iterrows():
                    graph_i = self.create_data_from_row(row)
                    data_train_list.append(graph_i)
                    if self.use_weights:
                        data_train_list_weights.append(row.weight)
                    progress.update(task, advance=1)
            # torch.save((data_train_list, data_train), "data.pkl")

            data_test_list = []
            with Progress() as progress:
                task = progress.add_task(
                    "Processing graphs (test)...", total=len(data_test)
                )

                for _, row in data_test.iterrows():
                    data_test_list.append(self.create_data_from_row(row))
                    progress.update(task, advance=1)
            # torch.save((data_test_list, data_test), "data_test.pkl")

            self.data_train = data_train_list
            self.data_test = data_test_list
            if self.use_weights:
                self.data_train_list_weights = data_train_list_weights

        else:
            self.data_train = data_train
            self.data_test = data_test
            self.data_test_raw = data_test

    def plot(
        self,
        filename,
        variables=None,
        overlay=None,
        bins=50,
        label=None,
    ):
        if overlay:
            if not isinstance(overlay, list):
                overlay = [overlay]

        if variables is None:
            variables = list(self.data.all_data["physical"].keys())

        with PdfPages(filename) as pdf:
            for variable in variables:
                try:
                    plt.figure(figsize=(10, 8))

                    plt.subplot(2, 2, 1)
                    plt.title(variable)
                    if overlay:
                        info = [self.data.all_data["physical"][variable]]
                        for overlay_i in overlay:
                            info.append(overlay_i.data.all_data["physical"][variable])
                        plt.hist(
                            info, bins=bins, density=True, histtype="step", label=label
                        )
                        # plt.legend()
                    else:
                        plt.hist(
                            self.data.all_data["physical"][variable],
                            bins=bins,
                            density=True,
                            histtype="step",
                            label=label,
                        )

                    if label:
                        plt.legend()

                    plt.subplot(2, 2, 2)
                    plt.title(f"{variable} processed")
                    if overlay:
                        info = [self.data.all_data["processed"][variable]]
                        for overlay_i in overlay:
                            info.append(overlay_i.data.all_data["processed"][variable])
                        plt.hist(info, bins=bins, density=True, histtype="step")
                    else:
                        plt.hist(
                            self.data.all_data["processed"][variable],
                            bins=bins,
                            density=True,
                            histtype="step",
                            range=[-1, 1],
                        )

                    plt.subplot(2, 2, 3)
                    if overlay:
                        info = [self.data.all_data["physical"][variable]]
                        for overlay_i in overlay:
                            info.append(overlay_i.all_data["physical"][variable])
                        plt.hist(info, bins=bins, density=True, histtype="step")
                    else:
                        plt.hist(
                            self.data.all_data["physical"][variable],
                            bins=bins,
                            density=True,
                            histtype="step",
                        )
                    plt.yscale("log")

                    plt.subplot(2, 2, 4)
                    if overlay:
                        info = [self.data.all_data["processed"][variable]]
                        for overlay_i in overlay:
                            info.append(overlay_i.data.all_data["processed"][variable])
                        plt.hist(info, bins=bins, density=True, histtype="step")
                    else:
                        plt.hist(
                            self.data.all_data["processed"][variable],
                            bins=bins,
                            density=True,
                            histtype="step",
                            range=[-1, 1],
                        )
                    plt.yscale("log")

                    pdf.savefig(bbox_inches="tight")
                    plt.close()

                except Exception:
                    pdf.savefig(bbox_inches="tight")
                    plt.close()
                    pass

    def save_graphs(self, save_loc):
        data = {
            "data_train": self.data_train,
            "data_test": self.data_test,
        }

        # Save the dictionary to a .pkl file
        with open(save_loc, "wb") as f:
            pickle.dump(data, f)

    def load_graphs(self, save_loc):
        # Load the data from the specified .pkl file
        with open(save_loc, "rb") as f:
            data = pickle.load(f)

        # Assign loaded data to the corresponding attributes
        self.data_train = data.get("data_train")
        self.data_test = data.get("data_test")

    def create_data_from_row(self, row):
        node_features = (
            torch.tensor(row[self.targets_node].values, dtype=torch.float).view(
                self.N_daughters, -1
            )
            # .to(myGlobals.device)
        )
        graph_features = torch.tensor(
            row[self.targets_graph].values, dtype=torch.float
        )  # .to(myGlobals.device)
        repeated_graph_features = graph_features.repeat(
            self.N_daughters, 1
        )  # Repeat for each node
        node_features_with_graph = torch.cat(
            [node_features, repeated_graph_features], dim=1
        )  # .to(myGlobals.device)

        node_conditions = torch.tensor(
            row[self.conditions_node].values, dtype=torch.float
        ).view(self.N_daughters, -1)
        graph_conditions = torch.tensor(
            row[self.conditions_graph].values, dtype=torch.float
        )
        repeated_graph_conditions = graph_conditions.repeat(
            self.N_daughters, 1
        )  # Repeat for each node
        node_conditions_with_graph = torch.cat(
            [node_conditions, repeated_graph_conditions], dim=1
        )  # .to(myGlobals.device)

        edge_index = []
        for i in range(self.N_daughters):
            for j in range(self.N_daughters):
                if j != i:  # no self-loops
                    # if i < j: # i think di-directionality is not implied so comment this out
                    edge_index.append([i, j])

        edge_index = (
            torch.tensor(edge_index, dtype=torch.long).t().contiguous()
            # .to(myGlobals.device)
        )

        # print(edge_index)
        # print(self.particles_involved)

        if len(self.conditions_edge) > 0:
            edge_features = []
            for i in self.particles_involved:
                for j in self.particles_involved:
                    if i != j:
                        # print(i, j)
                        edge_features.append(
                            [
                                row[f"edge_angle_DAUGHTER{i}_DAUGHTER{j}"],
                                row[f"edge_angle_DAUGHTER{i}_DAUGHTER{j}_TRUE"],
                            ]
                        )
        edge_features = torch.tensor(edge_features, dtype=torch.float)
        # print(edge_features)
        # quit()

        data = Data(
            edge_index=edge_index,
            targets=node_features_with_graph,  # inputs to each node, so need both graph and node features
            conditions=node_conditions_with_graph,  # inputs to each node, so need both graph and node features
            graph_targets=graph_features.unsqueeze(0),
            N_daughters=self.N_daughters,
            particles_involved=self.particles_involved,
        )
        if len(self.targets_node) > 0:
            data.node_targets = node_features
        if len(self.conditions_edge) > 0:
            data.edge_features = edge_features

        if not self.first_row_printed:
            print(data)
            self.first_row_printed = True

        return data

    def create_graphs(
        self,
        targets_graph,
        targets_node,
        conditions_graph,
        conditions_node,
        N_daughters_override=None,
        fully_reco_override=None,
    ):  # used by inference.py
        self.targets_graph = targets_graph
        self.targets_node = targets_node
        self.conditions_graph = conditions_graph
        self.conditions_node = conditions_node

        branches = (
            self.targets_graph
            + self.targets_node
            + self.conditions_graph
            + self.conditions_node
        )

        data_test = self.data.get_branches(branches)

        if N_daughters_override is not None:
            data_test["N_daughters"] = N_daughters_override
        if fully_reco_override is not None:
            data_test["fully_reco"] = fully_reco_override

        data_test_list = []
        with Progress() as progress:
            task = progress.add_task(
                "Processing graphs (test)...", total=len(data_test)
            )

            for _, row in data_test.iterrows():
                data_test_list.append(self.create_data_from_row(row))
                progress.update(task, advance=1)

        self.data_test = data_test_list
        self.data_test_raw = data_test

        return len(data_test_list)

    def get_loaders(
        self, batch_size, only_test=False, graphify=False
    ):  # used by inference.py
        if graphify:
            self.graphify = graphify

        if self.graphify:
            if not only_test:
                loader_train = geoDataLoader(
                    self.data_train, batch_size=batch_size, shuffle=True
                )
            test_batch_size = len(self.data_test)
            loader_test = geoDataLoader(
                self.data_test, batch_size=test_batch_size, shuffle=False
            )
        else:
            if not only_test:
                train_dataset = StandardDataset(
                    self.data_train, self.targets_graph, self.conditions_graph
                )
                loader_train = DataLoader(
                    train_dataset, batch_size=batch_size, shuffle=True
                )

            test_dataset = StandardDataset(
                self.data_test, self.targets_graph, self.conditions_graph
            )
            test_batch_size = len(test_dataset)
            loader_test = DataLoader(
                test_dataset, batch_size=test_batch_size, shuffle=False
            )
        if only_test:
            return loader_test
        else:
            return loader_train, loader_test, self.data_test_raw, test_batch_size



import lhcb_rex.settings.globals as myGlobals

import torch
import torch.nn as nn
from torch import Tensor
import pytorch_lightning as pl


def _weights_init(m):
    """Custom weight initialization."""
    if isinstance(m, nn.Linear):
        # Xavier initialization
        nn.init.xavier_normal_(m.weight)
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.0)
    elif isinstance(m, nn.BatchNorm1d):
        # BatchNorm weights should be initialized to 1 and biases to 0
        nn.init.constant_(m.weight, 1.0)
        nn.init.constant_(m.bias, 0.0)


class Discriminator(nn.Module):
    def __init__(self, targets_dims, conditions_dims, hidden_channels, latent_dims):
        super(Discriminator, self).__init__()

        if not isinstance(hidden_channels, list):
            hidden_channels = [hidden_channels, hidden_channels, hidden_channels]

        self.fc1 = nn.Linear(targets_dims + conditions_dims, hidden_channels[0])
        self.fc2 = nn.Linear(hidden_channels[0], hidden_channels[1])
        self.fc3 = nn.Linear(hidden_channels[1], hidden_channels[2])

        self.final_layer = nn.Linear(hidden_channels[2], 1)

        # Batch normalization layers
        self.bn1 = nn.BatchNorm1d(hidden_channels[0])
        self.bn2 = nn.BatchNorm1d(hidden_channels[1])
        self.bn3 = nn.BatchNorm1d(hidden_channels[2])

        self.dropout = nn.Dropout(0.2)

        self.apply(_weights_init)

    def forward(self, targets, conditions):
        x = torch.cat([targets, conditions], dim=1)

        # First hidden layer
        x = self.fc1(x)
        x = torch.nn.functional.leaky_relu(x)  # Activation function
        # x = self.bn1(x)  # Batch normalization
        # x = self.dropout(x)  # Dropout

        # Second hidden layer
        x = self.fc2(x)
        # x = self.bn2(x)  # Batch normalization
        x = torch.nn.functional.leaky_relu(x)  # Activation function
        # x = self.dropout(x)  # Dropout

        # Third hidden layer
        x = self.fc3(x)
        # x = self.bn3(x)  # Batch normalization
        x = torch.nn.functional.leaky_relu(x)  # Activation function
        # x = self.dropout(x)  # Dropout

        # Latent variables (mean and log variance for the VAE)
        disc = torch.sigmoid(self.final_layer(x))  # Mean of the latent space

        return disc


class Generator(pl.LightningModule):
    def __init__(self, targets_dims, conditions_dims, hidden_channels, latent_dims):
        super(Generator, self).__init__()

        if not isinstance(hidden_channels, list):
            hidden_channels = [hidden_channels, hidden_channels, hidden_channels]

        self.fc1 = nn.Linear(latent_dims + conditions_dims, hidden_channels[0])
        self.fc2 = nn.Linear(hidden_channels[0], hidden_channels[1])
        self.fc3 = nn.Linear(hidden_channels[1], hidden_channels[2])

        self.final_layer = nn.Linear(hidden_channels[2], targets_dims)

        # Batch normalization layers
        self.bn1 = nn.BatchNorm1d(hidden_channels[0])
        self.bn2 = nn.BatchNorm1d(hidden_channels[1])
        self.bn3 = nn.BatchNorm1d(hidden_channels[2])

        self.dropout = nn.Dropout(0.2)

        self.apply(_weights_init)

    # def forward(self, latent, conditions):
    def forward(
        self,
        latent: Tensor,
        conditions: Tensor,
    ) -> Tensor:
        x = torch.cat([latent, conditions], dim=1)

        # First hidden layer
        x = self.fc1(x)
        x = torch.nn.functional.leaky_relu(x)  # Activation function
        x = self.bn1(x)  # Batch normalization
        # x = self.dropout(x)  # Dropout

        # Second hidden layer
        x = self.fc2(x)
        x = torch.nn.functional.leaky_relu(x)  # Activation function
        x = self.bn2(x)  # Batch normalization
        # x = self.dropout(x)  # Dropout

        # Third hidden layer
        x = self.fc3(x)
        x = torch.nn.functional.leaky_relu(x)  # Activation function
        x = self.bn3(x)  # Batch normalization
        # x = self.dropout(x)  # Dropout

        # Latent variables (mean and log variance for the VAE)
        targets = torch.tanh(self.final_layer(x))  # Mean of the latent space

        return targets


class vanillaGAN(nn.Module):
    def __init__(
        self, targets_dims, conditions_dims, hidden_channels, latent_dims, silent=False
    ):
        super(vanillaGAN, self).__init__()

        if not silent:
            print("\n")
            print("Network params (vanillaGAN) -----------------")
            print("targets_dims:", targets_dims)
            print("conditions_dims:", conditions_dims)
            print("hidden_channels:", hidden_channels)
            print("latent_dims:", latent_dims)
            print("\n")

        self.latent_dims = latent_dims

        self.discriminator = Discriminator(
            targets_dims, conditions_dims, hidden_channels, latent_dims
        )

        self.generator = Generator(
            targets_dims, conditions_dims, hidden_channels, latent_dims
        )

    def inference(self, batch_size, conditions):
        # Sample latent variables for both graph and node using reparameterization trick
        z = torch.randn((batch_size, self.latent_dims)).to(myGlobals.device)
        conditions = conditions.to(myGlobals.device)
        self.generator = self.generator.to(myGlobals.device)

        # Decode the latent representation
        reconstructed = self.generator(z, conditions)

        return reconstructed



import lhcb_rex.settings.globals as myGlobals
from pydantic import BaseModel, validator, root_validator
import glob
import lhcb_rex.training.dataloader_handler as dh
import lhcb_rex.models.model_handler as model_handler
import lhcb_rex.training.hetero_trainer as trainer
import lhcb_rex.training.trainer as stdtrainer
from pytorch_lightning.loggers import TensorBoardLogger
from typing import Union, Any, Dict
from pytorch_lightning import Trainer
from lhcb_rex.data.data_handler import DataHandler
import torch_geometric
from pytorch_lightning.utilities.combined_loader import CombinedLoader


class FastSimTrainer(BaseModel):
    gpu: int = 5
    batch_size: int = 256
    use_weights: bool = False
    load_physics_validation: bool = True
    data_locs: Union[
        Union[str, Dict[str, str]],  # single string path
        Dict[str, Dict[str, Union[str, int]]],  # dict with path + batch_size
    ]

    model_parameters: str = "default_model_parameters.json"
    network: str
    logs_dir: str = "/users/am13743/Rex/logs"
    log_tag: str
    diffusion_model: str = None

    # # placeholders
    dataloaders: Any = None
    val_dataloaders: Any = None
    trainer: Any = None
    TrainerModule: Any = None
    logger: Any = None
    Model: Any = None
    extra_val_data_locs: Any = None

    @root_validator(pre=True)
    def validate_data_locs(cls, values):
        network = values.get("network", None)
        data_locs = values.get("data_locs", None)
        extra_val_data_locs = values.get("extra_val_data_locs", None)

        if "PV_smear" in network:
            if not isinstance(data_locs, str):
                raise ValueError(
                    f"For network={network} need data_locs to be a string (.root)"
                )
            if data_locs and not data_locs.endswith(".root"):
                raise ValueError(
                    f"For network={network} need data_locs to be a .root file"
                )
        else:
            if not isinstance(data_locs, dict):
                raise ValueError(f"For network={network} need data_locs to be a dict")

            for data_item, loc in data_locs.items():
                if isinstance(loc, dict):
                    loc = loc["loc"]
                if len(glob.glob(f"{loc}/*")) == 0:
                    raise ValueError(f"need a processed directory in {loc}")
                if glob.glob(f"{loc}/*")[0].split("/")[-1] != "processed":
                    raise ValueError(f"need a processed directory in {loc}")
                files = glob.glob(f"{loc}/processed/data_*.pt")
                files = [i for i in files if "_val_" not in i]
                if len(files) == 0:
                    raise ValueError(f"no data_*.pt files in {loc}/processed/")

        return values

    @validator("model_parameters")
    def validate_model_parameters(cls, value):
        if value and not value.endswith(".json"):
            raise ValueError("model_parameters must be a string ending in .json")
        return value

    @validator("network")
    def validate_network(cls, value):
        network_options = [
            "PV_smear",
            "PV_smear_diffusion",
            "mom_smear",
            "mom_smear_diffusion",
            "mom_smear_VAE",
            "PID_trig",
            "PID_trig_diffusion",
            "reco_vertex",
            "reco_vertex_diffusion",
            "reco_vertex_diffusion_eGAN",
            "mom_smear_flowmatching",
            "PID_trig_flowmatching",
            "reco_vertex_flowmatching",
        ]
        if value not in network_options:
            raise ValueError(f"network option must be one of: {network_options}")
        return value

    def __init__(self, **data):
        super().__init__(**data)

        print(f"Training on gpu {self.gpu}")

        if "PV_smear" in self.network:
            data_handler = DataHandler(
                particles_involved=[1],
                path=self.data_locs,
                N=-1,
                targets_graph=myGlobals.smearPV_targets,
                conditions_graph=myGlobals.smearPV_conditions,
                cut="abs(MOTHER_TRUEID)==521 and MOTHER_TRUEP_Z>0",
            )

            data_train, data_test, data_test_raw, test_batch_size = (
                data_handler.get_loaders(batch_size=self.batch_size)
            )

            branch_options = {}
            branch_options["targets_graph"] = myGlobals.smearPV_targets
            branch_options["conditions_graph"] = myGlobals.smearPV_conditions

            if "diffusion" in self.network:
                self.Model = model_handler.ModelHandler(
                    branch_options=branch_options,
                    graphify=False,
                    hidden_channels=[256, 512, 256],
                    latent_dims=len(myGlobals.smearPV_targets),
                    extra_latent_dims=3,
                    beta=80.0,
                    lr=0.0001,
                    network_option=self.network,
                )

            else:
                self.Model = model_handler.ModelHandler(
                    branch_options=branch_options,
                    graphify=False,
                    hidden_channels=[256, 512, 256],
                    latent_dims=25,
                    beta=80.0,
                    lr=0.0001,
                    network_option=self.network,
                )

            self.TrainerModule = stdtrainer.StdTrainerModule(
                self.Model,
                data_train,
                data_test,
                network_option=self.network,
            )

            self.dataloaders = data_train
            self.val_dataloaders = data_test

        else:
            settings = {}
            for data_item, loc_item in self.data_locs.items():
                if isinstance(loc_item, dict):
                    loc = loc_item["loc"]
                else:
                    loc = loc_item
                files = glob.glob(f"{loc}/processed/data_*.pt")
                files = [i for i in files if "_val_" not in i]
                files = [int(file.split("_")[-1].split(".")[0]) for file in files]
                # splits = len(files)
                splits = max(files) + 1
                if "batch_size" in loc_item:
                    settings[data_item] = dh.create_settings(
                        loc, splits=splits, batch_size=loc_item["batch_size"]
                    )
                else:
                    settings[data_item] = dh.create_settings(loc, splits=splits)

            if self.network == "reco_vertex":
                self.dataloaders, self.val_dataloaders, branch_options = dh.get_loaders(
                    settings,
                    batch_size=self.batch_size,
                    verbose=True,
                    use_weights=self.use_weights,
                    load_physics_validation=self.load_physics_validation,
                )
            elif self.network == "reco_vertex_diffusion":
                # enhance_fully_reco = 1.
                # enhance_fully_reco = 100.
                enhance_fully_reco = 250.0
                self.dataloaders, self.val_dataloaders, branch_options = dh.get_loaders(
                    settings,
                    batch_size=self.batch_size,
                    verbose=True,
                    use_weights=self.use_weights,
                    enhance_fully_reco=enhance_fully_reco,
                    extra_val_data_locs=self.extra_val_data_locs,
                )

                keys_to_keep = [
                    # 'N3_topIdx2',
                    "N3_topIdx1",
                    # 'N3_topIdx1_conditionless',
                    "N3_topIdx1_Kee_N3_topIdx1",
                ]
                self.val_dataloaders = {
                    k: self.val_dataloaders[k]
                    for k in keys_to_keep
                    if k in self.val_dataloaders
                }

            else:
                self.dataloaders, self.val_dataloaders, branch_options = dh.get_loaders(
                    settings,
                    batch_size=self.batch_size,
                    verbose=True,
                    use_weights=self.use_weights,
                )

            self.Model = model_handler.HeteroModelHandler(
                branch_options=branch_options,
                network_option=self.network,
                model_parameters=self.model_parameters,
                diffusion_model=self.diffusion_model,
            )

            if self.network == "reco_vertex_diffusion":
                self.TrainerModule = trainer.TrainerModule(
                    self.Model,
                    self.dataloaders,
                    self.val_dataloaders,
                    training_batch_size=self.batch_size,
                    save_models=True,
                    use_weights=self.use_weights,
                    network_option=self.network,
                    enhance_fully_reco=enhance_fully_reco,
                )
            else:
                self.TrainerModule = trainer.TrainerModule(
                    self.Model,
                    self.dataloaders,
                    self.val_dataloaders,
                    training_batch_size=self.batch_size,
                    save_models=True,
                    use_weights=self.use_weights,
                    network_option=self.network,
                )

        self.logger = TensorBoardLogger(self.logs_dir, name=self.log_tag)

    def fit(
        self,
        max_epochs=int(1e30),
        val_check_interval=None,
        test_validation=False,
        pre_load=None,
        cap_validation=-1,
        load_ema=False,
    ):
        self.trainer = Trainer(
            max_epochs=max_epochs,
            accelerator="gpu",
            devices=[self.gpu],
            logger=self.logger,
            val_check_interval=val_check_interval,
            enable_checkpointing=False,  # Disable automatic checkpoint saving - models saved in TrainerModule() at end of epochs (save_models=True)
            log_every_n_steps=5,
        )
        # min_size
        # max_size_cycle

        if pre_load:
            print(f"Loading {pre_load}...")
            self.Model.load(pre_load, ema=load_ema)

        if cap_validation > 0 and "PV_smear" not in self.network:
            for name, dataloader in self.val_dataloaders.items():
                dataset = dataloader.dataset
                num_samples = len(dataset)
                print(f"{name}: {num_samples} samples")
                if num_samples > cap_validation:
                    print(f"  -> Capping {name} at {cap_validation} samples")
                    # Truncate the dataset
                    truncated_dataset = dataset[:cap_validation]
                    # Re-create the dataloader with the truncated dataset
                    self.val_dataloaders[name] = torch_geometric.loader.DataLoader(
                        truncated_dataset,
                        batch_size=dataloader.batch_size,
                    )
            for name, dataloader in self.val_dataloaders.items():
                dataset = dataloader.dataset
                num_samples = len(dataset)
                print(f"{name}: {num_samples} samples")

        if test_validation:
            self.trainer.initialisation = False
            self.trainer.validate(self.TrainerModule, self.val_dataloaders)
            # self.trainer.validate(self.TrainerModule, self.dataloaders)
            if "smear" in self.network:
                self.TrainerModule.make_plots_smearingnet()
            else:
                self.TrainerModule.make_plots()
            self.TrainerModule.BDT_FoM()
            quit()

        self.dataloaders = CombinedLoader(self.dataloaders, mode="min_size")
        # self.dataloaders = CombinedLoader(self.dataloaders, mode='max_size_cycle')
        # min_size
        # max_size_cycle

        self.trainer.fit(
            self.TrainerModule, self.dataloaders, val_dataloaders=self.val_dataloaders
        )



import lhcb_rex.settings.globals as myGlobals
from pydantic import BaseModel, validator, StrictStr, ValidationError
from typing import Optional
import re
from contextlib import contextmanager
import inspect
import traceback


@contextmanager
def execute_command(kwargs, param_class):
    caller_function = inspect.stack()[-2].function

    try:
        params = param_class(**kwargs)  # Use the passed parameter class
        yield params.dict()  # Pass the validated params to the `with` block
    except ValidationError as ve:
        # Handle Pydantic validation errors
        myGlobals.console.print(
            f"[red3][bold]Validation error in {caller_function}()[/bold][/red3]"
        )
        for error in ve.errors():
            myGlobals.console.print(
                f"[red3][bold]{error['loc'][0]}[/bold][/red3] - {error['msg']}"
            )
    except Exception as e:
        # Handle general exceptions and print the stack trace
        myGlobals.console.print(
            f"[red3][bold]Error in {caller_function}()[/bold][/red3]"
        )
        myGlobals.console.print(f"[red3][bold]Exception:[/bold][/red3] {str(e)}")
        myGlobals.console.print(
            f"[red3][bold]Traceback:[/bold][/red3]\n{traceback.format_exc()}"
        )
        raise


class lhcbsim_runFromTupleParameters(BaseModel):
    tuple_location: StrictStr
    mother_particle: StrictStr
    daughter_particles: list
    reconstruction_topology: Optional[StrictStr] = None
    mass_hypotheses: Optional[dict] = None
    intermediate_particles: Optional[dict] = None
    branch_naming_structure: Optional[dict] = None
    physical_units: StrictStr = "MeV"
    stages: list[StrictStr] = ["runVertexing"]

    @validator("branch_naming_structure")
    def check_decay_format(cls, v):
        rapidsim_options = [
            "momenta_component",
            "true_momenta_component",
            "momenta_component",
            "true_momenta_component",
            "pid",
            "true_pid",
            "mass",
            "true_mass",
            "origin",
            "true_origin",
            "vertex",
            "true_vertex",
        ]
        for key in v:
            if key not in rapidsim_options:
                raise ValueError(
                    f"'{v}' is not a valid option; must be one of {rapidsim_options}."
                )
        return v

    @validator("reconstruction_topology")
    def check_reconstruction_topology_format(cls, v):
        if "->" not in v:
            raise ValueError(f"'{v}' is invalid; it must contain '->'.")
        mothers = v.split("->", 1)[0].split()  # split only once
        num_mothers = len([mother for mother in mothers if mother])
        if num_mothers != 1:
            raise ValueError(
                f"'{v}' is invalid; must define exactly one mother particle."
            )
        return v

    @validator("mass_hypotheses")
    def check_mass_hypotheses(cls, v, values):
        if v:
            for particle_name in v:
                if particle_name not in values.get("daughter_particles"):
                    raise ValueError(
                        f"'{particle_name}' is invalid; not present in daughter_particles '{values.get('daughter_particles')}'."
                    )
        return v

    @validator("intermediate_particles")
    def check_intermediate_particle(cls, v, values):
        if v:
            for intermediate_name, particle_names in v.items():
                for particle_name in particle_names:
                    if particle_name not in values.get("daughter_particles"):
                        raise ValueError(
                            f"'{intermediate_name}' cannot be built; '{particle_name}' not present in daughter_particles '{values.get('daughter_particles')}'."
                        )
        return v

    @validator("physical_units")
    def check_physical_units(cls, v):
        units_options = {"MeV", "GeV"}
        if v not in units_options:
            raise ValueError(
                f"'{v}' is not a valid option; must be one of {units_options}."
            )
        return v


class lhcbsim_runParameters(BaseModel):
    events: int
    decay: StrictStr
    naming_scheme: StrictStr
    decay_models: Optional[StrictStr] = None
    mass_hypotheses: Optional[dict] = None
    intermediate_particles: Optional[dict] = None
    reconstruction_topology: Optional[StrictStr] = None
    geometry: StrictStr = "LHCb"
    acceptance: StrictStr = "AllIn"
    useEvtGen: StrictStr = "TRUE"
    evtGenUsePHOTOS: StrictStr = "TRUE"
    dropMissing: bool = True
    verbose: bool = False
    only_rapidsim: bool = False
    workingDir: StrictStr = "./decay"
    clean_up_files: bool = True
    keep_conditions: bool = False
    run_systematics: bool = False

    @validator("useEvtGen", "evtGenUsePHOTOS")
    def check_boolean_strings(cls, value):
        normalized_value = value.strip().upper()
        if normalized_value not in {"TRUE", "FALSE"}:
            raise ValueError(f"'{value}' must be either 'TRUE' or 'FALSE'")
        return normalized_value

    @validator("decay", "naming_scheme", "decay_models")
    def remove_double_spaces(cls, v):
        return re.sub(r"\s+", " ", v).strip()

    @validator("reconstruction_topology")
    def check_reconstruction_topology_format(cls, v):
        if "->" not in v:
            raise ValueError(f"'{v}' is invalid; it must contain '->'.")
        mothers = v.split("->", 1)[0].split()  # split only once
        num_mothers = len([mother for mother in mothers if mother])
        if num_mothers != 1:
            raise ValueError(
                f"'{v}' is invalid; must define exactly one mother particle."
            )
        return v

    @validator("decay")
    def check_decay_format(cls, v):
        if "->" not in v:
            raise ValueError(f"'{v}' is invalid; it must contain '->'.")
        mothers = v.split("->", 1)[0].split()  # split only once
        num_mothers = len([mother for mother in mothers if mother])
        if num_mothers != 1:
            raise ValueError(
                f"'{v}' is invalid; must define exactly one mother particle."
            )
        return v

    @validator("naming_scheme")
    def check_naming_scheme_structure(cls, v, values):
        decay = values.get("decay")
        if decay is None:
            raise ValueError(
                "Decay structure must be provided before validating naming_scheme."
            )
        decay_tokens = decay.split()
        if v:
            if "->" not in v:
                raise ValueError(f"'{v}' is invalid; it must contain '->'.")
            v_tokens = v.split()
            if len(v_tokens) != len(decay_tokens):
                raise ValueError(
                    f"'{v}' is invalid; it must match the structure of decay '{decay}' (replace any spaces in individual items with underscores)."
                )
        else:
            # Automatically generate naming scheme if not provided
            v = " ".join(
                "PHSP" if token not in {"{", "}", "->"} else token
                for token in decay_tokens
            )
        return v

    @validator("decay_models")
    def check_naming_scheme_structure2(cls, v, values):
        decay = values.get("decay")
        if decay is None:
            raise ValueError(
                "Decay structure must be provided before validating decay_models."
            )

        decay_tokens = decay.split()

        def tokenize(s):
            # Match either [[...]] blocks or normal tokens
            pattern = r"\[\[.*?\]\]|\S+"
            return re.findall(pattern, s)

        v_tokens = tokenize(v)

        if "->" not in v_tokens:
            raise ValueError(f"'{v}' is invalid; it must contain '->'.")

        if len(v_tokens) != len(decay_tokens):
            raise ValueError(
                f"'{v}' is invalid; it must match the structure of decay '{decay}' "
                f"(wrap compound models in [[ ]] if needed)."
            )

        return v

    @validator("naming_scheme", allow_reuse=True)
    def check_naming_scheme(cls, v):
        tokens = v.split()
        for token in tokens:
            if token not in {"{", "}", "->"} and len(token) <= 1:
                raise ValueError(
                    f"Particle names must be longer than 1 character, offending name: '{token}'."
                )
        return v

    @validator("mass_hypotheses")
    def check_mass_hypotheses(cls, v, values):
        naming_scheme = values.get("naming_scheme")
        if naming_scheme and v:
            valid_particles = set(naming_scheme.split())
            for particle_name in v:
                if particle_name not in valid_particles:
                    raise ValueError(
                        f"'{particle_name}' is invalid; not present in naming_scheme '{naming_scheme}' (you might need to pad out particles with spaces)."
                    )
        return v

    @validator("intermediate_particles")
    def check_intermediate_particles(cls, v, values):
        naming_scheme = values.get("naming_scheme")
        if naming_scheme and v:
            valid_particles = set(naming_scheme.split())
            for intermediate_name, particle_names in v.items():
                for particle_name in particle_names:
                    if particle_name not in valid_particles:
                        raise ValueError(
                            f"'{intermediate_name}' cannot be built; '{particle_name}' not present in naming_scheme '{naming_scheme}' (you might need to pad out particles with spaces)."
                        )
        return v

    @validator("geometry")
    def check_geometry(cls, v):
        rapidsim_options = {"4pi", "LHCb"}
        if v not in rapidsim_options:
            raise ValueError(
                f"'{v}' is not a valid option; must be one of {rapidsim_options}."
            )
        return v

    @validator("acceptance")
    def check_acceptance(cls, v):
        rapidsim_options = {"Any", "ParentIn", "AllIn", "AllDownstream"}
        if v not in rapidsim_options:
            raise ValueError(
                f"'{v}' is not a valid option; must be one of {rapidsim_options}."
            )
        return v


class fvqi_runFromTupleParameters(BaseModel):
    file: StrictStr
    mother_particle: StrictStr
    daughter_particles: list
    fully_reco: bool
    nPositive_missing_particles: int
    nNegative_missing_particles: int
    mass_hypotheses: Optional[dict] = None
    intermediate_particle: Optional[dict] = None
    branch_naming_structure: Optional[dict] = None
    physical_units: StrictStr = "GeV"
    keep_conditions: bool = False

    @validator("branch_naming_structure")
    def check_decay_format(cls, v):
        rapidsim_options = [
            "momenta_component",
            "true_momenta_component",
            "momenta_component",
            "true_momenta_component",
            "pid",
            "true_pid",
            "mass",
            "true_mass",
            "origin",
            "true_origin",
            "vertex",
            "true_vertex",
        ]
        for key in v:
            if key not in rapidsim_options:
                raise ValueError(
                    f"'{v}' is not a valid option; must be one of {rapidsim_options}."
                )
        return v

    @validator("mass_hypotheses")
    def check_mass_hypotheses(cls, v, values):
        if v:
            for particle_name in v:
                if particle_name not in values.get("daughter_particles"):
                    raise ValueError(
                        f"'{particle_name}' is invalid; not present in daughter_particles '{values.get('daughter_particles')}'."
                    )
        return v

    @validator("intermediate_particle")
    def check_intermediate_particle(cls, v, values):
        if v:
            for intermediate_name, particle_names in v.items():
                for particle_name in particle_names:
                    if particle_name not in values.get("daughter_particles"):
                        raise ValueError(
                            f"'{intermediate_name}' cannot be built; '{particle_name}' not present in daughter_particles '{values.get('daughter_particles')}'."
                        )
        return v

    @validator("physical_units")
    def check_physical_units(cls, v):
        units_options = {"MeV", "GeV"}
        if v not in units_options:
            raise ValueError(
                f"'{v}' is not a valid option; must be one of {units_options}."
            )
        return v


class fvqi_runParameters(BaseModel):
    events: int
    decay: StrictStr
    naming_scheme: StrictStr
    decay_models: Optional[StrictStr] = None
    mass_hypotheses: Optional[dict] = None
    intermediate_particle: Optional[dict] = None
    geometry: StrictStr = "LHCb"
    acceptance: StrictStr = "AllIn"
    useEvtGen: StrictStr = "TRUE"
    evtGenUsePHOTOS: StrictStr = "TRUE"
    dropMissing: bool = True
    verbose: bool = False
    only_rapidsim: bool = False
    workingDir: StrictStr = "./decay"
    clean_up_files: bool = True
    keep_conditions: bool = False
    run_systematics: bool = False

    @validator("useEvtGen", "evtGenUsePHOTOS")
    def check_boolean_strings(cls, value):
        normalized_value = value.strip().upper()
        if normalized_value not in {"TRUE", "FALSE"}:
            raise ValueError(f"'{value}' must be either 'TRUE' or 'FALSE'")
        return normalized_value

    @validator("decay", "naming_scheme", "decay_models")
    def remove_double_spaces(cls, v):
        return re.sub(r"\s+", " ", v).strip()

    @validator("decay")
    def check_decay_format(cls, v):
        if "->" not in v:
            raise ValueError(f"'{v}' is invalid; it must contain '->'.")
        mothers = v.split("->", 1)[0].split()  # split only once
        num_mothers = len([mother for mother in mothers if mother])
        if num_mothers != 1:
            raise ValueError(
                f"'{v}' is invalid; must define exactly one mother particle."
            )
        return v

    # @validator("naming_scheme", "decay_models")
    # def check_naming_scheme_structure(cls, v, values):
    #     decay = values.get("decay")
    #     if decay is None:
    #         raise ValueError(
    #             "Decay structure must be provided before validating naming_scheme."
    #         )
    #     decay_tokens = decay.split()
    #     if v:
    #         if "->" not in v:
    #             raise ValueError(f"'{v}' is invalid; it must contain '->'.")
    #         v_tokens = v.split()
    #         if len(v_tokens) != len(decay_tokens):
    #             raise ValueError(
    #                 f"'{v}' is invalid; it must match the structure of decay '{decay}' (replace any spaces in individual items with underscores)."
    #             )
    #     else:
    #         # Automatically generate naming scheme if not provided
    #         v = " ".join(
    #             "PHSP" if token not in {"{", "}", "->"} else token
    #             for token in decay_tokens
    #         )
    #     return v

    @validator("naming_scheme", allow_reuse=True)
    def check_naming_scheme(cls, v):
        tokens = v.split()
        for token in tokens:
            if token not in {"{", "}", "->"} and len(token) <= 1:
                raise ValueError(
                    f"Particle names must be longer than 1 character, offending name: '{token}'."
                )
        return v

    @validator("mass_hypotheses")
    def check_mass_hypotheses(cls, v, values):
        naming_scheme = values.get("naming_scheme")
        if naming_scheme and v:
            valid_particles = set(naming_scheme.split())
            for particle_name in v:
                if particle_name not in valid_particles:
                    raise ValueError(
                        f"'{particle_name}' is invalid; not present in naming_scheme '{naming_scheme}' (you might need to pad out particles with spaces)."
                    )
        return v

    @validator("intermediate_particle")
    def check_intermediate_particle(cls, v, values):
        naming_scheme = values.get("naming_scheme")
        if naming_scheme and v:
            valid_particles = set(naming_scheme.split())
            for intermediate_name, particle_names in v.items():
                for particle_name in particle_names:
                    if particle_name not in valid_particles:
                        raise ValueError(
                            f"'{intermediate_name}' cannot be built; '{particle_name}' not present in naming_scheme '{naming_scheme}' (you might need to pad out particles with spaces)."
                        )
        return v

    @validator("geometry")
    def check_geometry(cls, v):
        rapidsim_options = {"4pi", "LHCb"}
        if v not in rapidsim_options:
            raise ValueError(
                f"'{v}' is not a valid option; must be one of {rapidsim_options}."
            )
        return v

    @validator("acceptance")
    def check_acceptance(cls, v):
        rapidsim_options = {"Any", "ParentIn", "AllIn", "AllDownstream"}
        if v not in rapidsim_options:
            raise ValueError(
                f"'{v}' is not a valid option; must be one of {rapidsim_options}."
            )
        return v



import lhcb_rex


def test_basic():
    lhcb_rex.run(
        events=250,
        decay="B+ -> K+ e+ e-",
        naming_scheme="MOTHER -> DAUGHTER1 DAUGHTER2 DAUGHTER3",
        decay_models="BTOSLLBALL_6 -> NA NA NA",
        reconstruction_topology="MOTHER -> DAUGHTER1 { INTERMEDIATE -> DAUGHTER2 DAUGHTER3 }",
        workingDir="./pytest",
    )



import lhcb_rex.settings.globals as myGlobals
from lhcb_rex.rapidsim.run_rapidsim import run_rapidsim
from lhcb_rex.inference.run_network import run_inference_from_tuple as run_network
import lhcb_rex.tools.display as display

import lhcb_rex.inference.params as p


def run_from_tuple(**kwargs):
    with p.execute_command(kwargs, p.lhcbsim_runFromTupleParameters) as params:

        # Run network with rapidsim output
        rapidsim_tuple_reco = run_network(
            params["tuple_location"],
            mother_particle=params["mother_particle"],
            daughter_particles=params["daughter_particles"],
            intermediate_particles=params["intermediate_particles"],
            branch_naming_structure=params["branch_naming_structure"],
            reconstruction_topology=params["reconstruction_topology"],
            stages=params["stages"],  # dont run smearing networks
            physical_units=params["physical_units"],
            mass_hypotheses=params["mass_hypotheses"],
        )

        # Display timing and file information
        total_time = display.timings_table(only_vertexing=True)
        display.print_file_info(rapidsim_tuple_reco, time=total_time)


def run(**kwargs):
    with p.execute_command(kwargs, p.lhcbsim_runParameters) as params:
        
        # Set verbosity if needed
        if params["verbose"]:
            myGlobals._verbose = True

        # Unpack the required parameters for run_rapidsim and run_network
        rapidsim_output = run_rapidsim(
            params["workingDir"],
            params["events"],
            params["decay"],
            params["naming_scheme"],
            params["decay_models"],
            params["mass_hypotheses"],
            params["intermediate_particles"],
            params["geometry"],
            params["acceptance"],
            params["useEvtGen"],
            params["evtGenUsePHOTOS"],
            params["dropMissing"],
            params["clean_up_files"],
        )

        (
            rapidsim_tuple,
            fully_reco,
            nPositive_missing_particles,
            nNegative_missing_particles,
            mother_particle,
            daughter_particles,
            true_PID_scheme,
            combined_particles,
            map_NA_codes,
        ) = rapidsim_output


        # Display events table
        display.events_table(params["events"], rapidsim_tuple)

        if params["only_rapidsim"]:
            total_time = display.timings_table(only_rapidsim=params["only_rapidsim"])
            display.print_file_info(rapidsim_tuple, time=total_time)
            return

        intermediate_particles = combined_particles.copy()
        del intermediate_particles["MOTHER"]

        rapidsim_tuple_reco = run_network(
            rapidsim_tuple,
            mother_particle=mother_particle,
            daughter_particles=daughter_particles,
            intermediate_particles=intermediate_particles,
            reconstruction_topology=params["reconstruction_topology"],
            stages=[
                "smearPV",
                "smearMomenta",
                "genPID",
                "runVertexing",
            ],  # dont run smearing networks
            physical_units="GeV",  # RapidSim
            mass_hypotheses=params["mass_hypotheses"],
        )

        # Display timing and file information
        total_time = display.timings_table()
        display.print_file_info(rapidsim_tuple_reco, time=total_time)



import lhcb_rex.settings.globals as myGlobals

import pytorch_lightning as pl
import torch
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import pandas as pd
from tensorboard.backend.event_processing import event_accumulator
import numpy as np
import torch.nn.functional as F
from torchmetrics.regression import MeanAbsolutePercentageError
import torch.nn as nn
from scipy.stats import norm
from diffusers import DDPMScheduler

mean_abs_percentage_error = MeanAbsolutePercentageError()


_EPSILON = 1e-7  # Small constant for numerical stability


class StdTrainerModule(pl.LightningModule):
    def __init__(self, model, dataloaders, val_dataloaders, network_option="PV_smear"):
        super(StdTrainerModule, self).__init__()

        self.dataloaders = dataloaders
        self.val_dataloaders = val_dataloaders
        self.models = model
        self.network_option = network_option

        self.targets_graph = model.branch_options["targets_graph"]

        self.gan = self.models.gan
        self.automatic_optimization = False

        self.learning_rate = self.models.lr

        self.beta = self.models.beta

        self.initialisation = True

    def _loss_generator(self, y_pred):
        # Clip predictions to avoid log(0)
        y_pred = torch.clamp(y_pred, _EPSILON, 1.0 - _EPSILON)
        # Calculate loss
        out = -torch.log(y_pred)
        # Return mean loss
        return torch.mean(out, dim=-1)

    def training_step(self, batch, batch_idx):
        if "diffusion" in self.network_option:
            optimizer = self.optimizers()
            self.gan.diffuser.train()

            device = batch["targets"].device

            batch_size_i = batch["targets"].shape[0]

            timesteps = torch.randint(
                low=0,
                high=self.gan.noise_scheduler.config.num_train_timesteps,
                size=(batch_size_i,),
                device=device,
            )

            # Add noise
            noise = self.gan.sample_starting_noise(
                batch["targets"].shape, device=device
            )
            noisy_batch = self.gan.noise_scheduler.add_noise(
                batch["targets"], noise, timesteps
            )

            # Gen latent
            latent = torch.randn(
                (batch_size_i, self.gan.latent_dims),
                dtype=noisy_batch.dtype,
                device=device,
            )

            current_state = torch.cat((noisy_batch, latent), dim=1)

            noise_pred = self.gan.diffuser(
                current_state, batch["conditions"], timesteps
            )

            loss = F.l1_loss(noise, noise_pred, reduction="mean")  # or mse?

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            self.gan.diffuser.eval()

            self.log("loss", loss, prog_bar=True)

        else:
            optimizer_g, optimizer_d = self.optimizers()
            self.gan.discriminator.train()
            self.gan.generator.eval()

            device = batch["targets"].device
            self.gan.generator.to(device)
            self.gan.discriminator.to(device)

            batch_size_i = int(batch["targets"].shape[0] / 3)

            sample_targets_train_B = batch["targets"][batch_size_i : 2 * batch_size_i]
            sample_conditions_train_A = batch["conditions"][:batch_size_i]
            sample_conditions_train_B = batch["conditions"][
                batch_size_i : 2 * batch_size_i
            ]
            sample_conditions_train_C = batch["conditions"][
                2 * batch_size_i : 3 * batch_size_i
            ]

            # Generate noise and conditions
            noise = torch.randn(batch_size_i, self.models.latent_dims).to(device)
            generated_images = self.gan.generator(noise, sample_conditions_train_A)

            # Concatenate inputs and labels for the discriminator
            in_values = torch.cat([generated_images, sample_targets_train_B], dim=0)
            in_values_labels = torch.cat(
                [sample_conditions_train_A, sample_conditions_train_B], dim=0
            )

            labels_D_0 = torch.zeros(batch_size_i, 1).to(device)
            labels_D_1 = torch.ones(batch_size_i, 1).to(device)
            labels_D = torch.cat([labels_D_0, labels_D_1], dim=0)

            # Discriminator loss
            out_values_choice = self.gan.discriminator(in_values, in_values_labels)
            disc_loss_fn = nn.BCELoss()

            disc_loss = disc_loss_fn(out_values_choice.squeeze(), labels_D.squeeze())

            # Backpropagation for discriminator
            optimizer_d.zero_grad()
            disc_loss.backward()
            optimizer_d.step()

            self.gan.discriminator.eval()
            self.gan.generator.train()

            # Generate noise and conditions for the generator
            noise_stacked = torch.randn(batch_size_i, self.models.latent_dims).to(
                device
            )

            # Generator loss
            fake_images2 = self.gan.generator(noise_stacked, sample_conditions_train_C)
            stacked_output_choice = self.gan.discriminator(
                fake_images2, sample_conditions_train_C
            )
            gen_loss = self._loss_generator(stacked_output_choice.squeeze())

            # Backpropagation for generator
            optimizer_g.zero_grad()
            gen_loss.backward()
            optimizer_g.step()

            self.gan.discriminator.eval()
            self.gan.generator.eval()

            self.log("disc_loss", disc_loss)
            self.log("gen_loss", gen_loss)

    def run_inference(self, batch, test_batch_size):
        reconstructed_graph = self.models.gan.inference(
            test_batch_size,
            batch["conditions"],
        )

        reconstructed_graph = reconstructed_graph.cpu().detach().numpy()

        gen_df_i = pd.DataFrame(
            reconstructed_graph,
            columns=self.targets_graph,
        )

        return gen_df_i

    def create_true_df(self, batch, test_batch_size):
        graph_targets = batch["targets"].cpu().detach().numpy()

        true_df_i = pd.DataFrame(
            graph_targets,
            columns=self.targets_graph,
        )

        return true_df_i

    def on_validation_epoch_start(self):
        print(f"Starting validation for epoch {self.current_epoch}")

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        test_batch_size = batch["targets"].shape[0]

        gen_df_i = self.run_inference(batch, test_batch_size)

        if batch_idx == 0:
            self.gen_data = gen_df_i
        else:
            self.gen_data = pd.concat((self.gen_data, gen_df_i), axis=0)

        if self.current_epoch == 0:  # TRUE info
            true_df_i = self.create_true_df(batch, test_batch_size)

            if batch_idx == 0:
                self.true_data = true_df_i
            else:
                self.true_data = pd.concat((self.true_data, true_df_i), axis=0)

    def make_plots(self):
        print("Plotting...")
        pdf_name = f"{self.logger.log_dir}/epoch{self.current_epoch}.pdf"

        with PdfPages(pdf_name) as pdf:
            for branch in self.gen_data:
                gen_df = self.gen_data
                # reco_df = self.reco_data
                true_df = self.true_data

                plt.title(branch)
                plt.hist(
                    [gen_df[branch], true_df[branch]],
                    bins=35,
                    density=True,
                    histtype="step",
                    label=["gen", "true"],
                    range=[-1, 1],
                )
                plt.legend()
                pdf.savefig(bbox_inches="tight")
                plt.close()

                plt.title(branch)
                plt.hist2d(
                    gen_df[branch],
                    true_df[branch],
                    bins=35,
                    range=[[-1, 1], [-1, 1]],
                )
                plt.xlabel("gen")
                plt.ylabel("true")
                pdf.savefig(bbox_inches="tight")
                plt.close()

        print(f"Plotted {pdf_name}.")

    def load(self, models_loc):
        print("Loading models...")
        self.models.load(models_loc)

    def on_validation_epoch_end(self):
        if self.initialisation:  # no need to run this stuff for the initialistion step.
            self.initialisation = False
            return

        print("Saving models...")
        self.models.save(f"{self.logger.log_dir}/models.pkl")

        # Make plots
        self.make_plots()

    def configure_optimizers(self):
        if "diffusion" in self.network_option:
            opt_g = torch.optim.AdamW(
                self.gan.diffuser.parameters(),
                lr=1e-4,
                betas=(0.9, 0.999),
                amsgrad=False,
            )

            return [opt_g], []

        else:
            opt_g = torch.optim.Adam(
                self.gan.generator.parameters(),
                lr=0.0004,
                betas=(0.5, 0.999),
                amsgrad=True,
            )
            opt_d = torch.optim.Adam(
                self.gan.discriminator.parameters(),
                lr=0.0004,
                betas=(0.5, 0.999),
                amsgrad=True,
            )

            return [opt_g, opt_d], []



import lhcb_rex.settings.globals as myGlobals

import torch
import torch.nn as nn
from torch import Tensor
import pytorch_lightning as pl
from diffusers import DDPMScheduler


def _weights_init(m):
    """Custom weight initialization."""
    if isinstance(m, nn.Linear):
        # Xavier initialization
        nn.init.xavier_normal_(m.weight)
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.0)
    elif isinstance(m, nn.BatchNorm1d):
        # BatchNorm weights should be initialized to 1 and biases to 0
        nn.init.constant_(m.weight, 1.0)
        nn.init.constant_(m.bias, 0.0)


class TimeEmbedding(nn.Module):
    def __init__(self, embedding_dim):
        super().__init__()
        self.embedding_dim = embedding_dim
        self.linear1 = nn.Linear(1, embedding_dim)
        self.act = nn.SiLU()
        self.linear2 = nn.Linear(embedding_dim, embedding_dim)

    def forward(self, t):
        # t is shape [batch_size], scale to float and unsqueeze
        t = t.unsqueeze(-1).float()
        emb = self.linear1(t)
        emb = self.act(emb)
        emb = self.linear2(emb)
        return emb


class Diffuser(pl.LightningModule):
    def __init__(self, targets_dims, conditions_dims, hidden_channels, latent_dims):
        super(Diffuser, self).__init__()

        self.time_embedding_dim = 5
        self.time_embed = TimeEmbedding(self.time_embedding_dim)

        if not isinstance(hidden_channels, list):
            hidden_channels = [hidden_channels, hidden_channels, hidden_channels]

        self.fc1 = nn.Linear(
            targets_dims + latent_dims + conditions_dims + self.time_embedding_dim,
            hidden_channels[0],
        )
        self.fc2 = nn.Linear(hidden_channels[0], hidden_channels[1])
        self.fc3 = nn.Linear(hidden_channels[1], hidden_channels[2])

        self.final_layer = nn.Linear(hidden_channels[2], targets_dims)

        # Batch normalization layers
        self.bn1 = nn.BatchNorm1d(hidden_channels[0])
        self.bn2 = nn.BatchNorm1d(hidden_channels[1])
        self.bn3 = nn.BatchNorm1d(hidden_channels[2])

        self.dropout = nn.Dropout(0.2)

    # def forward(self, latent, conditions):
    def forward(
        self,
        batch,
        conditions,
        timesteps,
    ) -> Tensor:
        timestep_embedding = self.time_embed(timesteps)

        x = torch.cat([batch, conditions, timestep_embedding], dim=1)

        # First hidden layer
        x = self.fc1(x)
        x = torch.nn.functional.leaky_relu(x)  # Activation function
        x = self.bn1(x)  # Batch normalization
        x = self.dropout(x)  # Dropout

        # Second hidden layer
        x = self.fc2(x)
        x = torch.nn.functional.leaky_relu(x)  # Activation function
        x = self.bn2(x)  # Batch normalization
        x = self.dropout(x)  # Dropout

        # Third hidden layer
        x = self.fc3(x)
        x = torch.nn.functional.leaky_relu(x)  # Activation function
        x = self.bn3(x)  # Batch normalization
        x = self.dropout(x)  # Dropout

        targets = self.final_layer(x)

        return targets


class vanillaDiffusion(nn.Module):
    def __init__(
        self, targets_dims, conditions_dims, hidden_channels, latent_dims, silent=False
    ):
        super(vanillaDiffusion, self).__init__()

        if not silent:
            print("\n")
            print("Network params (vanillaDiffusion) -----------------")
            print("targets_dims:", targets_dims)
            print("conditions_dims:", conditions_dims)
            print("hidden_channels:", hidden_channels)
            print("latent_dims:", latent_dims)
            print("\n")

        self.latent_dims = latent_dims

        # self.noise_scheduler = DDPMScheduler(num_train_timesteps=model.gan.steps, clip_sample=True, clip_sample_range=5.5, beta_schedule="squaredcos_cap_v2")
        self.noise_scheduler = DDPMScheduler(
            num_train_timesteps=25, beta_schedule="squaredcos_cap_v2"
        )

        self.targets_dims = targets_dims
        self.diffuser = Diffuser(
            targets_dims, conditions_dims, hidden_channels, latent_dims
        )

    def sample_starting_noise(self, shape, device):
        """
        I think this might have to be random normal 0,1
        """
        return torch.randn(shape, device=device)

    def inference(self, test_batch_size, batch: torch.Tensor) -> torch.Tensor:
        device = batch.device
        current_state = self.sample_starting_noise(
            (batch.shape[0], self.targets_dims), device=device
        )
        latent = torch.randn((batch.shape[0], self.latent_dims), device=device)
        conditions = batch

        for t in reversed(range(self.noise_scheduler.config.num_train_timesteps)):
            timesteps = torch.full(
                (batch.shape[0],), t, device=device, dtype=torch.long
            )

            batch_i = torch.cat((current_state, latent), dim=1)
            noise_pred = self.diffuser(
                batch=batch_i,
                conditions=conditions,
                timesteps=timesteps,
            )

            current_state = self.noise_scheduler.step(
                model_output=noise_pred, timestep=t, sample=current_state
            ).prev_sample

        return current_state
